from .arena_object import Object

class Triangle(Object):
    """
    Triangle object class to manage its properties in the ARENA: Triangle Geometry.
    
    :param dict vertexA: vertex A Defaults to '{'x': 0, 'y': 0.5, 'z': 0}' (optional)
    :param dict vertexB: vertex B Defaults to '{'x': -0.5, 'y': -0.5, 'z': 0}' (optional)
    :param dict vertexC: vertex C Defaults to '{'x': 0.5, 'y': -0.5, 'z': 0}' (optional)
    """
    object_type = "triangle"

    def __init__(self, **kwargs):
        super().__init__(object_type=Triangle.object_type, **kwargs)



from .attribute import Attribute

class Material(Attribute):
    """
    Material attribute class to manage its properties in the ARENA: The material properties of the object's surface.  More properties at (https://aframe.io/docs/1.5.0/components/material.html) A-Frame Material.
    Usage: material=Material(...)
    
    :param float alphaTest: Alpha test threshold for transparency. (optional)
    :param str blending: The blending mode for the material's RGB and Alpha sent to the WebGLRenderer. Allows [none, normal, additive, subtractive, multiply] Defaults to 'normal' (optional)
    :param str color: Base diffuse color. Defaults to '#7f7f7f' (optional)
    :param bool depthTest: Whether depth testing is enabled when rendering the material. Defaults to 'True' (optional)
    :param bool dithering: Whether material is dithered with noise. Removes banding from gradients like ones produced by lighting. Defaults to 'True' (optional)
    :param bool flatShading: Use THREE.FlatShading rather than THREE.StandardShading. (optional)
    :param bool npot: Use settings for non-power-of-two (NPOT) texture. (optional)
    :param dict offset: Texture offset to be used. Defaults to '{'x': 1, 'y': 1}' (optional)
    :param float opacity: Extent of transparency. If the transparent property is not true, then the material will remain opaque and opacity will only affect color. Defaults to '1' (optional)
    :param dict repeat: Texture repeat to be used. Defaults to '{'x': 1, 'y': 1}' (optional)
    :param str shader: Which material to use. Defaults to the standard material. Can be set to the flat material or to a registered custom shader material. Defaults to 'standard' (optional)
    :param str side: Which sides of the mesh to render. Allows [front, back, double] Defaults to 'front' (optional)
    :param str src: URI, relative or full path of an image/video file. e.g. 'store/users/wiselab/images/360falls.mp4'. (optional)
    :param bool transparent: Whether material is transparent. Transparent entities are rendered after non-transparent entities. (optional)
    :param str vertexColors: Whether to use vertex or face colors to shade the material. Allows [none, vertex, face] Defaults to 'none' (optional)
    :param bool visible: Whether material is visible. Raycasters will ignore invisible materials. Defaults to 'True' (optional)
    """
    def __init__(self, **kwargs):
        if "opacity" in kwargs:
            # kwargs["transparent"] = True # need to be transparent to be opaque
            if kwargs["opacity"] > 1.0: # keep opacity between 0.0 and 1.0
                kwargs["opacity"] = float(kwargs["opacity"]) / 100
            kwargs["opacity"] = max(0.0, kwargs["opacity"])
            kwargs["opacity"] = min(kwargs["opacity"], 1.0)

        super().__init__(**kwargs)



# This file is auto-generated from github.com/arenaxr/arena-schema, changes here may be overwritten.

from .animation import *
from .animation_mixer import *
from .armarker import *
from .attribution import *
from .blip import *
from .box_collision_listener import *
from .click_listener import *
from .dynamic_body import *
from .gltf_model_lod import *
from .gltf_morph import *
from .goto_landmark import *
from .goto_url import *
from .impulse import *
from .jitsi_video import *
from .landmark import *
from .material import *
from .material_extras import *
from .model_update import *
from .multisrc import *
from .position import *
from .remote_render import *
from .rotation import *
from .scale import *
from .shadow import *
from .sound import *
from .spe_particles import *
from .static_body import *
from .textinput import *
from .video_control import *

ATTRIBUTE_KEYWORD_TRANSLATION = {
    "physics": "dynamic-body",  # backward-compatibility
    "clickable": "click-listener",  # backward-compatibility
    "animation": "animation",
    "animation_mixer": "animation-mixer",
    "armarker": "armarker",
    "attribution": "attribution",
    "blip": "blip",
    "box_collision_listener": "box-collision-listener",
    "buffer": "buffer",
    "click_listener": "click-listener",
    "collision_listener": "collision-listener",
    "dynamic_body": "dynamic-body",
    "gltf_model_lod": "gltf-model-lod",
    "gltf_morph": "gltf-morph",
    "goto_landmark": "goto-landmark",
    "goto_url": "goto-url",
    "hide_on_enter_ar": "hide-on-enter-ar",
    "hide_on_enter_vr": "hide-on-enter-vr",
    "impulse": "impulse",
    "jitsi_video": "jitsi-video",
    "landmark": "landmark",
    "look_at": "look-at",
    "material": "material",
    "material_extras": "material-extras",
    "modelUpdate": "modelUpdate",
    "multisrc": "multisrc",
    "parent": "parent",
    "position": "position",
    "remote_render": "remote-render",
    "rotation": "rotation",
    "scale": "scale",
    "screenshareable": "screenshareable",
    "shadow": "shadow",
    "show_on_enter_ar": "show-on-enter-ar",
    "show_on_enter_vr": "show-on-enter-vr",
    "skipCache": "skipCache",
    "sound": "sound",
    "spe_particles": "spe-particles",
    "static_body": "static-body",
    "textinput": "textinput",
    "url": "url",
    "video_control": "video-control",
}

KEYWORD_ATTRIBUTE_TRANSLATION = {
    "animation": "animation",
    "animation-mixer": "animation_mixer",
    "armarker": "armarker",
    "attribution": "attribution",
    "blip": "blip",
    "box-collision-listener": "box_collision_listener",
    "buffer": "buffer",
    "click-listener": "click_listener",
    "collision-listener": "collision_listener",
    "dynamic-body": "dynamic_body",
    "gltf-model-lod": "gltf_model_lod",
    "gltf-morph": "gltf_morph",
    "goto-landmark": "goto_landmark",
    "goto-url": "goto_url",
    "hide-on-enter-ar": "hide_on_enter_ar",
    "hide-on-enter-vr": "hide_on_enter_vr",
    "impulse": "impulse",
    "jitsi-video": "jitsi_video",
    "landmark": "landmark",
    "look-at": "look_at",
    "material": "material",
    "material-extras": "material_extras",
    "modelUpdate": "modelUpdate",
    "multisrc": "multisrc",
    "parent": "parent",
    "position": "position",
    "remote-render": "remote_render",
    "rotation": "rotation",
    "scale": "scale",
    "screenshareable": "screenshareable",
    "shadow": "shadow",
    "show-on-enter-ar": "show_on_enter_ar",
    "show-on-enter-vr": "show_on_enter_vr",
    "skipCache": "skipCache",
    "sound": "sound",
    "spe-particles": "spe_particles",
    "static-body": "static_body",
    "textinput": "textinput",
    "url": "url",
    "video-control": "video_control",
}

ATTRIBUTE_CLASS_TRANSLATION = {
    "animation": Animation,
    "animation_mixer": AnimationMixer,
    "armarker": Armarker,
    "attribution": Attribution,
    "blip": Blip,
    "box_collision_listener": BoxCollisionListener,
    "click_listener": ClickListener,
    "dynamic_body": DynamicBody,
    "gltf_model_lod": GltfModelLod,
    "gltf_morph": GltfMorph,
    "goto_landmark": GotoLandmark,
    "goto_url": GotoUrl,
    "impulse": Impulse,
    "jitsi_video": JitsiVideo,
    "landmark": Landmark,
    "material": Material,
    "material_extras": MaterialExtras,
    "modelUpdate": ModelUpdate,
    "multisrc": Multisrc,
    "position": Position,
    "remote_render": RemoteRender,
    "rotation": Rotation,
    "scale": Scale,
    "shadow": Shadow,
    "sound": Sound,
    "spe_particles": SpeParticles,
    "static_body": StaticBody,
    "textinput": Textinput,
    "video_control": VideoControl,
}



from .attribute import Attribute


class Attribution(Attribute):
    """
    Attribution attribute class to manage its properties in the ARENA: Attribution Component. Saves attribution data in any entity.
    Usage: attribution=Attribution(...)
    
    :param str author: Author name; e.g. 'Vaptor-Studio'. Defaults to 'Unknown' (optional)
    :param str authorURL: Author homepage/profile; e.g. https://sketchfab.com/VapTor. (optional)
    :param bool extractAssetExtras: Extract attribution info from asset extras; will override attribution info given (default: true). Defaults to 'True' (optional)
    :param str license: License summary/short name; e.g. 'CC-BY-4.0'. Defaults to 'Unknown' (optional)
    :param str licenseURL: License URL; e.g. http://creativecommons.org/licenses/by/4.0/. (optional)
    :param str source: Model source e.g. 'Sketchfab'. Defaults to 'Unknown' (optional)
    :param str sourceURL: Model source URL; e.g. https://sketchfab.com/models/2135501583704537907645bf723685e7. (optional)
    :param str title: Model title; e.g. 'Spinosaurus'. Defaults to 'No Title' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class BoxCollisionListener(Attribute):
    """
    BoxCollisionListener attribute class to manage its properties in the ARENA: Listen for bounding-box collisions with user camera and hands. Must be applied to an object or model with geometric mesh. Collisions are determined by course bounding-box overlaps.
    Usage: box_collision_listener=BoxCollisionListener(...)
    
    :param bool dynamic: Set true for a moving object, which should have its bounding box recalculated regularly to determine proper collision. (optional)
    :param bool enabled: Publish detections, set `false` to disable. Defaults to 'True' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class Shadow(Attribute):
    """
    Shadow attribute class to manage its properties in the ARENA: 
    Usage: shadow=Shadow(...)
    
    :param bool cast: Whether the entity casts shadows onto the surrounding scene. (optional)
    :param bool receive: Whether the entity receives shadows from the surrounding scene. (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object

class Torus(Object):
    """
    Torus object class to manage its properties in the ARENA: Torus Geometry.
    
    :param float arc: Arc Defaults to '360' (optional)
    :param float radius: radius Defaults to '1' (optional)
    :param float radiusTubular: radius tubular Defaults to '0.2' (optional)
    :param int segmentsRadial: segments radial Defaults to '36' (optional)
    :param int segmentsTubular: segments tubular Defaults to '32' (optional)
    """
    object_type = "torus"

    def __init__(self, **kwargs):
        super().__init__(object_type=Torus.object_type, **kwargs)



"""
The ArenaTelemetry generates traces, metrics, and logs using OpenTelemetry (OTEL).
It can export using OTEL's protocol (OTLP), send JSON OTEL spans to MQTT, or to the console. 

The :envvar:`ARENA_TELEMETRY` environment variable enables the telemetry.
The :envvar:`OTLP_ENDPOINT` environment variable defines the OTLP endpoint when OTLP is used.
"""
import atexit
import sys
import os
import signal 
import time

from typing import Callable, Optional, List, Sequence

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider, ReadableSpan, Span
from opentelemetry.trace import NoOpTracerProvider, Status, StatusCode
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter, SpanExporter, SpanExportResult
from opentelemetry.trace.span import INVALID_SPAN

from ..env import (
    ARENA_TELEMETRY,
    OTLP_ENDPOINT,
    _get_env
)

TRACE_TOPIC_DFT = "realm/ns/scene/t/traces"

class MQTTSpanExporter(SpanExporter):
    """Implementation of :class:`SpanExporter` that sends spans to MQTT"""

    def __init__(
        self,
        topic: Optional[str] = TRACE_TOPIC_DFT,
        service_name: Optional[str] = None,
        formatter: Callable[
            [ReadableSpan], str
        ] = lambda span: span.to_json()
        + os.linesep,
    ):
        # TODO: create mqtt client and assign its publish calls
        self.publish = None
        self.publish_flush = None
        
        self.topic = topic
        self.formatter = formatter
        self.service_name = service_name

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            #self.out.write(self.formatter(span))
            self.publish(self.topic, self.formatter(span))
        if self.publish_flush: self.publish_flush(True)
        return SpanExportResult.SUCCESS

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return True
    
    def shutdown(self) -> None:
        print("shutdown exporter!")
        pass
    
class ArenaTelemetry():
    """Implementation of ARENA telemetry. 
       According to :envvar:`ARENA_TELEMETRY`, exports using OTLP, send JSON OTEL spans to MQTT, or to the console. 
    """
    
    parent_span: Span = None

    def __init__(self, name=sys.argv[0], id=None):
        """ Return an `ArenaTelemetry` using given service name and id
            Provides utility calls that wrap open telemetry functionality to start spans, log events, and other.
        
            Creates a parent span for all the spans related to the program.
        
            Args:
                name: name of the service used with the telemetry backend
                id: additional id used with the telemetry backend
        """
            
        service_name = f"{name}"
        if id: service_name = service_name + "({id})"
        # Service name is required for most backends
        resource = Resource(attributes={
            SERVICE_NAME: service_name
        })
        
        env_telemetry = _get_env(ARENA_TELEMETRY)
        otlp_endpoint = _get_env(OTLP_ENDPOINT)
        tel_exporters = {
            'otlp': lambda: OTLPSpanExporter(otlp_endpoint, insecure=True),
            'mqtt': lambda: MQTTSpanExporter(), 
            'console': lambda: ConsoleSpanExporter()
        }
        
        if env_telemetry: self.enabled = env_telemetry.lower() in tel_exporters.keys()
        else: 
            env_telemetry = 'none'
            self.enabled = False
        
        if self.enabled:
            provider = TracerProvider(resource=resource)
            self.exporter = tel_exporters[env_telemetry]()
            processor = BatchSpanProcessor(self.exporter)
            provider.add_span_processor(processor)
            trace.set_tracer_provider(provider)
        else:
            if not env_telemetry.lower() in ('none', 'false', ''): print(f"Warn: Invalid telemetry processor specified: {env_telemetry}")
            # instanciate a no-op tracer
            provider = NoOpTracerProvider()
            trace.set_tracer_provider(provider)

        # creates a tracer from the global tracer provider
        self.tracer = trace.get_tracer("ARENA-py")
        
        # create a parent span
        self.parent_span = self.tracer.start_span("somescene")
        self.parent_span_ctx = trace.set_span_in_context(self.parent_span)    
        
        # make sure we end parent span
        atexit.register(self.exit) 
             
    def exit(self, error_msg=None):
        """Record exit status on error """
        if not self.enabled: return
        if error_msg: self.parent_span.set_status(Status(StatusCode.ERROR, error_msg))
        self.parent_span.end()
        self.exporter.force_flush()
        self.parent_span = INVALID_SPAN
        
    def __del__(self):
        if self.parent_span != INVALID_SPAN: self.exit()
        
    def start_span(self, name, **kwargs):
        """Wrapper to otel start_as_current_span; force context to be parent span"""
        if 'context' in kwargs: del kwargs['context']
        return self.tracer.start_as_current_span(name, context=self.parent_span_ctx, **kwargs)

    def start_process_msg_span(self, obj_id, action, **kwargs):
        """Wrapper to otel start_as_current_span to start a process message span; force context to be parent span"""
        if 'context' in kwargs: del kwargs['context']
        return self.tracer.start_as_current_span(f"{obj_id} {action} process_message", context=self.parent_span_ctx, **kwargs)

    def start_publish_span(self, obj_id, action, type, **kwargs):
        """Wrapper to otel start_as_current_span to start a process message span; force context to be parent span"""
        if 'context' in kwargs: del kwargs['context']
        return self.tracer.start_as_current_span(f"{obj_id} {action} publish_message {type}", context=self.parent_span_ctx, **kwargs)
        
    def add_event(self, name, span=None, print_msg=True, **kwargs):
        """Add event to given or current span"""
        if print_msg: print(name)
        if not self.enabled: return
        if not span: span = trace.get_current_span()
        if span == INVALID_SPAN: span = self.parent_span
        span.add_event(name, kwargs)

    def set_error(self, error_msg, span=None, print_msg=True):
        """Set error on given or current span"""
        if print_msg: print(error_msg)
        if not self.enabled: return
        if not span: span = trace.get_current_span()
        if span == INVALID_SPAN: span = self.parent_span
        span.set_status(Status(StatusCode.ERROR, error_msg))



from .attribute import Attribute


class StaticBody(Attribute):
    """
    StaticBody attribute class to manage its properties in the ARENA: A fixed-position or animated object. Other objects may collide with static bodies, but static bodies themselves are unaffected by gravity and collisions.  More properties at (https://github.com/c-frame/aframe-physics-system/blob/master/CannonDriver.md) A-Frame Physics System.
    Usage: static_body=StaticBody(...)
    
    :param str cylinderAxis: Override default axis of bounding cylinder. Requires shape:cylinder. Allows [x, y, z] Defaults to 'y' (optional)
    :param str shape: Body components will attempt to find an appropriate CANNON.js shape to fit your model. When defining an object you may choose a shape or leave the default, auto. Select a shape carefully, as there are performance implications with different choices. Allows [auto, box, cylinder, sphere, hull, none] Defaults to 'auto' (optional)
    :param float sphereRadius: Override default radius of bounding sphere. Requires shape:sphere. NaN by default. (optional)
    :param str type: Define the result of collisions. Dynamic can be moved, Static cannot be moved. Allows [dynamic, static] Defaults to 'static' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class ClickListener(Attribute):
    """
    ClickListener attribute class to manage its properties in the ARENA: Object will listen for clicks.
    Usage: click_listener=ClickListener(...) or clickable=ClickListener(...)  or clickable=True

    :param bool bubble: Set false to prevent click events from bubbling up to parent objects. See <a href='https://developer.mozilla.org/en-US/docs/Learn/JavaScript/Building_blocks/Events#event_bubbling'>Event Bubbling</a>. Defaults to 'True' (optional)
    :param bool enabled: Publish events, set false to disable. Defaults to 'True' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .attribute import Attribute


class MaterialExtras(Attribute):
    """
    MaterialExtras attribute class to manage its properties in the ARENA: Define extra material properties, namely texture encoding, whether to render the material's color and render order.  The properties set here access directly Three.js material component.   More properties at (https://threejs.org/docs/#api/en/materials/Material) THREE.js Material.
    Usage: material_extras=MaterialExtras(...)
    
    :param bool colorWrite: Whether to render the material's color. Defaults to 'True' (optional)
    :param str encoding:  Allows [LinearEncoding, sRGBEncoding, GammaEncoding, RGBEEncoding, LogLuvEncoding, RGBM7Encoding, RGBM16Encoding, RGBDEncoding, BasicDepthPacking, RGBADepthPacking] Defaults to 'sRGBEncoding' (optional)
    :param float gltfOpacity: Opacity value to apply to the model. 1 is fully opaque, 0 is fully transparent. Defaults to '1' (optional)
    :param str overrideSrc: Overrides the material source in all meshes of an object (e.g. a basic shape or a GLTF); Use, for example, to change the texture of a GLTF. (optional)
    :param float renderOrder: Allows the default rendering order of scene graph objects to be overridden. Defaults to '1' (optional)
    :param bool transparentOccluder: If `true`, will set `colorWrite=false` and `renderOrder=0` to make the material a transparent occluder. (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



""" 
Collect information about the program execution, such as 
activity times and number of messages sent/received.
"""

from datetime import datetime
import os
import sys

from ..event_loop import PersistentWorker
from ..base_object import BaseObject
from ..env import (
    PROGRAM_STATS_UPDATE_INTERVAL_MS,
    _get_env,
    _get_arena_env
)

class GetPublicAttrsMixin():
    def get_attrs(self, **kwargs):
        """ Return object public members only """              
        obj = {k: v for k, v in vars(self).items() if k.startswith("_") == False}
        obj.update(kwargs)
        return obj
        
class QueueStats(dict):
    def __init__(self, rcv_queue_len, pub_queue_len):
        self.rcv_queue_len = rcv_queue_len
        self.pub_queue_len = pub_queue_len
    
class ProgramRunInfo(BaseObject, GetPublicAttrsMixin):
    """Program Run Information; collect program execution information. """
        
    object_type = "run_info"

    def __init__(self, evt_loop=None, queue_len_callable=None, update_callback=None, update_interval_ms=_get_env(PROGRAM_STATS_UPDATE_INTERVAL_MS), **kwargs):
        """
        Returns a `ProgramRunInfo`. If an event loop is passed, will setup a periodic task to 
        update execution stats and perform a callback to notify of this update
        
        Args:
            evt_loop: an event loop to which we add a periodic task to update program stats
            queue_len_callable: callable that returns a queue stats object
            update_callback: callback when stats are updated
            update_interval_ms: interval of the periodic task to update program stats
            kwargs: additional attributes to add to the program info, e.g. web host, scene, namespace
        """
        # add additional arguments 
        for key, value in kwargs.items():
            setattr(self, key, value)
        
        # program args, env
        self.filename=sys.argv[0]
        self.args=sys.argv[1:]
        self.env=_get_arena_env()
        
        # run stats
        self.create_time =  datetime.utcnow().isoformat()[:-3]+"Z"
        self.last_active_time =  datetime.utcnow().isoformat()[:-3]+"Z"
        self.rcv_msgs = 0
        self.pub_msgs = 0
        # init when a message is sent/received (so we can see if messages were never sent/received)
        self.last_rcv_time = None
        self.last_pub_time = None
        # init when avg is computed
        self.avg_rcv_msgs_per_sec = None
        self.avg_pub_msgs_per_sec = None
        
        self._update_callback = update_callback
        self._get_queue_len = queue_len_callable            

        self._msg_rate_time_start = datetime.now()
        self._rcv_msgs_start = self.rcv_msgs
        self._pub_msgs_start = self.rcv_msgs
        self._update_count = 0
        
        if evt_loop:
            # update stats periodically if event loop is given
            t = PersistentWorker(evt_loop, self._update_stats, interval=update_interval_ms)
            evt_loop.add_task(t)
    
    def _update_stats(self):
        """Update stats (called by periodic task); Execute callback if defined """
        elapsed = datetime.now() - self._msg_rate_time_start
        if elapsed.seconds < 1: 
            return
        
        # compute moving avg of rcv and pub messages
        N=5 # window
        rcv_msgs = self.rcv_msgs - self._rcv_msgs_start
        pub_msgs = self.pub_msgs - self._pub_msgs_start
        self._rcv_msgs_start=self.rcv_msgs
        self._pub_msgs_start=self.rcv_msgs
        if elapsed.seconds > 0:
            rcv_msgs_per_sec = rcv_msgs  / elapsed.seconds
            pub_msgs_per_sec = pub_msgs  / elapsed.seconds
            if self._update_count < N: 
                if self.avg_rcv_msgs_per_sec == None:
                    self.avg_rcv_msgs_per_sec = 0
                    self.avg_pub_msgs_per_sec = 0
                self._update_count = self._update_count + 1
                N=self._update_count
            self.avg_rcv_msgs_per_sec = round(self.avg_rcv_msgs_per_sec + (rcv_msgs_per_sec - self.avg_rcv_msgs_per_sec) / N, 2)
            self.avg_pub_msgs_per_sec = round(self.avg_pub_msgs_per_sec + (pub_msgs_per_sec - self.avg_pub_msgs_per_sec) / N, 2)
                    
            # update queue lens if callable given
            if self._get_queue_len:
                for key, value in self._get_queue_len().items(): setattr(self, key, value)
                
            # update callback if given
            if self._update_callback: 
                self._update_callback(self)
        
    def msg_rcv(self):
        """Update stats when a message is received """
        self.last_rcv_time = datetime.utcnow().isoformat()[:-3]+"Z"
        self.rcv_msgs = self.rcv_msgs + 1
        self.last_active_time = self.last_rcv_time

    def msg_publish(self):
        """Update stats when a message is published """
        self.last_pub_time = datetime.utcnow().isoformat()[:-3]+"Z"
        self.pub_msgs = self.pub_msgs + 1
        self.last_active_time = self.last_pub_time
            
    def add_program_info(self, adict):
        """ Add program info to another dictionary """
        adict[ProgramRunInfo.object_type] = self.get_attrs()
        return adict
        


from ..utils import *
from .attribute import Attribute
from .position import Position
from .rotation import Rotation
from .goto_url import GotoUrl
from .jitsi_video import JitsiVideo
from .dynamic_body import Physics
from .scale import Scale
from .material import Material
from .color import Color
from .video_control import VideoControl
from .translate import ATTRIBUTE_CLASS_TRANSLATION, KEYWORD_ATTRIBUTE_TRANSLATION

class Data(Attribute):
    """
    Data attribute class to manage its properties in the ARENA: Wraps all attributes in JSON.
    Usage: data=Data(...)
    
    :param dict animation: Animate and tween values. (optional)
    :param dict animation_mixer: A list of available animations can usually be found by inspecting the model file or its documentation. All animations will play by default. To play only a specific set of animations, use wildcards: animation-mixer='clip: run_*'. (optional)
    :param dict armarker: A location marker (such as an AprilTag, a lightAnchor, or an UWB tag), used to anchor scenes, or scene objects, in the real world. (optional)
    :param dict attribution: Attribution Component. Saves attribution data in any entity. (optional)
    :param dict blip: When the object is created or deleted, it will animate in/out of the scene instead of appearing/disappearing instantly. Must have a geometric mesh. (optional)
    :param dict box_collision_listener: Listen for bounding-box collisions with user camera and hands. Must be applied to an object or model with geometric mesh. Collisions are determined by course bounding-box overlaps. (optional)
    :param bool buffer: Transform geometry into a BufferGeometry to reduce memory usage at the cost of being harder to manipulate (geometries only: box, circle, cone, ...). Defaults to 'True' (optional)
    :param dict click_listener: Object will listen for clicks. (optional)
    :param str collision_listener: Name of the collision-listener, default can be empty string. Collisions trigger click events. (optional)
    :param dict dynamic_body: A freely-moving object. Dynamic bodies have mass, collide with other objects, bounce or slow during collisions, and fall if gravity is enabled. (optional)
    :param dict gltf_model_lod: Simple switch between the default gltf-model and a detailed one when a user camera is within specified distance (optional)
    :param dict gltf_morph: Allows you to target and control a gltf model's morphTargets created in Blender. (optional)
    :param dict goto_landmark: Teleports user to the landmark with the given name; Requires click-listener. (optional)
    :param dict goto_url: Goto given URL; Requires click-listener. (optional)
    :param bool hide_on_enter_ar: Hide object when entering AR. Remove component to *not* hide. (optional)
    :param bool hide_on_enter_vr: Hide object when entering VR. Remove component to *not* hide. (optional)
    :param dict impulse: Apply an impulse to an object to set it in motion. This happens in conjunction with an event. Requires click-listener and physics. (optional)
    :param dict jitsi_video: Apply a jitsi video source to the geometry. (optional)
    :param dict landmark: Define entities as a landmark; Landmarks appears in the landmark list and you can move (teleport) to them; You can define the behavior of the teleport: if you will be at a fixed or random distance, looking at the landmark, fixed offset or if it is constrained by a navmesh (when it exists). (optional)
    :param str look_at: The look-at component defines the behavior for an entity to dynamically rotate or face towards another entity or position. Use '#my-camera' to face the user camera, otherwise can take either a vec3 position or a query selector to another entity. (optional)
    :param dict material: The material properties of the object's surface. (optional)
    :param dict material_extras: Define extra material properties, namely texture encoding, whether to render the material's color and render order. (optional)
    :param dict modelUpdate: The GLTF-specific `modelUpdate` attribute is an object with child component names as keys. The top-level keys are the names of the child components to be updated. The values of each are nested `position` and `rotation` attributes to set as new values, respectively. Either `position` or `rotation` can be omitted if unchanged. (optional)
    :param dict multisrc: Define multiple visual sources applied to an object. (optional)
    :param str parent: Parent's object_id. Child objects inherit attributes of their parent, for example scale and translation. (optional)
    :param dict position: 3D object position. (optional)
    :param dict remote_render: Whether or not an object should be remote rendered [Experimental]. (optional)
    :param dict rotation: 3D object rotation in quaternion representation; Right-handed coordinate system. Euler degrees are deprecated in wire message format. (optional)
    :param dict scale: 3D object scale. (optional)
    :param bool screenshareable: Whether or not a user can screenshare on an object. Defaults to 'True' (optional)
    :param dict shadow: Shadow (optional)
    :param bool show_on_enter_ar: Show object when entering AR. Hidden otherwise. (optional)
    :param bool show_on_enter_vr: Show object when entering VR. Hidden otherwise. (optional)
    :param bool skipCache: Disable retrieving the shared geometry object from the cache. (geometries only: box, circle, cone, ...). (optional)
    :param dict sound: The sound component defines the entity as a source of sound or audio. The sound component is positional and is thus affected by the component's position. (optional)
    :param dict spe_particles: GPU based particle systems in A-Frame. (optional)
    :param dict static_body: A fixed-position or animated object. Other objects may collide with static bodies, but static bodies themselves are unaffected by gravity and collisions. (optional)
    :param dict textinput: Opens an HTML prompt when clicked. Sends text input as an event on MQTT. Requires click-listener. (optional)
    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    :param dict video_control: Video Control. (optional)
    """
    def __init__(self, **kwargs):
        data = {}
        data = Data.update_data(data, kwargs)
        super().__init__(**data)

    @classmethod
    def update_data(cls, data, new_data):
        new_data = new_data.get("data", new_data)
        dash_words = []
        for k,v in new_data.items():
            # allow user to input tuples, lists, dicts, etc for specific Attributes.
            # everything gets converted to corresponding attribute
            if (k == "position" or k == "start" or k == "end") and not isinstance(v, Position):
                if isinstance(v, (list,tuple)):
                    data[k] = Position(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Position(**v)
                else:
                    data[k] = v

            elif k == "rotation" and not isinstance(v, Rotation):
                if isinstance(v, (list,tuple)):
                    if len(v) == 3:
                        data[k] = Rotation(*v[:3], None)
                    else:
                        data[k] = Rotation(*v[:4])
                elif isinstance(v, dict):
                    data[k] = Rotation(**v)
                else:
                    data[k] = v

            elif k == "scale" and not isinstance(v, Scale):
                if isinstance(v, (list,tuple)):
                    data[k] = Scale(*v[:3])
                elif isinstance(v, dict):
                    data[k] = Scale(**v)
                else:
                    data[k] = v

            elif k == "color" and not isinstance(v, Color):
                if isinstance(v, (list,tuple)):
                    color = Color(*v[:3])
                elif isinstance(v, dict):
                    color = Color(**v)
                elif isinstance(v, str):
                    color = Color(v)
                else:
                    color = v
                data[k] = color

            elif k == "material":
                if "color" in v:
                    color = v["color"]
                    if isinstance(color, (list,tuple)):
                        color = Color(*color[:3])
                    elif isinstance(color, dict):
                        color = Color(**color)
                    elif isinstance(color, str):
                        color = Color(color)
                    else:
                        color = v["color"]
                    v["color"] = color
                if isinstance(v, dict):
                    data[k] = Material(**v)
                else:
                    data[k] = v

            # Translate and handle underscores from any other keys.
            # Must be done last since KEYWORD_ATTRIBUTE_TRANSLATION contains all attributes,
            # which may interfere with special casing above from, say "rotation" for example.
            elif k in KEYWORD_ATTRIBUTE_TRANSLATION:
                if '-' in k:
                    dash_words += [k]
                    k = KEYWORD_ATTRIBUTE_TRANSLATION[k]
                if isinstance(v, dict):
                    data[k] = ATTRIBUTE_CLASS_TRANSLATION[k](**v)
                else:
                    data[k] = v

            elif isinstance(v, Attribute):
                data[k] = v

            else:
                try:
                    # unknown attribute
                    data[k] = Attribute(**v)
                except:
                    data[k] = v

            # make False into None
            if (isinstance(v, bool) and v == False) or v is None:
                data[k] = None

        # delete elements with keys that have dashes
        for w in dash_words:
            if w in data:
                del data[w]

        return data



# This file is auto-generated from github.com/arenaxr/arena-schema, changes here may be overwritten.

from .arena_object import *
from .arenaui_button_panel import *
from .arenaui_card import *
from .arenaui_prompt import *
from .box import *
from .camera import *
from .capsule import *
from .circle import *
from .cone import *
from .cylinder import *
from .dodecahedron import *
from .entity import *
from .gaussian_splatting import *
from .gltf_model import *
from .hand_left import *
from .hand_right import *
from .icosahedron import *
from .image import *
from .light import *
from .line import *
from .ocean import *
from .octahedron import *
from .pcd_model import *
from .plane import *
from .ring import *
from .roundedbox import *
from .sphere import *
from .tetrahedron import *
from .text import *
from .thickline import *
from .threejs_scene import *
from .torus import *
from .torus_knot import *
from .triangle import *
from .videosphere import *
from .program import *

OBJECT_TYPE_MAP = {
    Object.object_type: Object,
    ArenauiButtonPanel.object_type: ArenauiButtonPanel,
    ArenauiCard.object_type: ArenauiCard,
    ArenauiPrompt.object_type: ArenauiPrompt,
    Box.object_type: Box,
    Camera.object_type: Camera,
    Capsule.object_type: Capsule,
    Circle.object_type: Circle,
    Cone.object_type: Cone,
    Cylinder.object_type: Cylinder,
    Dodecahedron.object_type: Dodecahedron,
    Entity.object_type: Entity,
    GaussianSplatting.object_type: GaussianSplatting,
    GltfModel.object_type: GltfModel,
    HandLeft.object_type: HandLeft,
    HandRight.object_type: HandRight,
    Icosahedron.object_type: Icosahedron,
    Image.object_type: Image,
    Light.object_type: Light,
    Line.object_type: Line,
    Ocean.object_type: Ocean,
    Octahedron.object_type: Octahedron,
    PcdModel.object_type: PcdModel,
    Plane.object_type: Plane,
    Ring.object_type: Ring,
    Roundedbox.object_type: Roundedbox,
    Sphere.object_type: Sphere,
    Tetrahedron.object_type: Tetrahedron,
    Text.object_type: Text,
    Thickline.object_type: Thickline,
    ThreejsScene.object_type: ThreejsScene,
    Torus.object_type: Torus,
    TorusKnot.object_type: TorusKnot,
    Triangle.object_type: Triangle,
    Videosphere.object_type: Videosphere,
}


from .animation import Animation

class AnimationMixer(Animation):
    """
    AnimationMixer attribute class to manage its properties in the ARENA: A list of available animations can usually be found by inspecting the model file or its documentation. All animations will play by default. To play only a specific set of animations, use wildcards: animation-mixer='clip: run_*'.  More properties at (https://github.com/n5ro/aframe-extras/tree/master/src/loaders#animation) A-Frame Extras Animation.
    Usage: animation_mixer=AnimationMixer(...)
    
    :param bool clampWhenFinished: If true, halts the animation at the last frame. (optional)
    :param str clip: Name of the animation clip(s) to play. Accepts wildcards. Defaults to '*' (optional)
    :param float crossFadeDuration: Duration of cross-fades between clips, in seconds. (optional)
    :param float duration: Duration of the animation, in seconds (0 = auto). (optional)
    :param str loop: In repeat and pingpong modes, the clip plays once plus the specified number of repetitions. For pingpong, every second clip plays in reverse. Allows [once, repeat, pingpong] Defaults to 'repeat' (optional)
    :param str repetitions: Number of times to play the clip, in addition to the first play (empty string = Infinity). Repetitions are ignored for loop: once. (optional)
    :param float startAt: Sets the start of an animation to a specific time (in milliseconds). This is useful when you need to jump to an exact time in an animation. The input parameter will be scaled by the mixer's timeScale. (optional)
    :param float timeScale: Scaling factor for playback speed. A value of 0 causes the animation to pause. Negative values cause the animation to play backwards. Defaults to '1' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import math
from ..utils import Utils
from .attribute import Attribute
from collections.abc import Iterable


class Position(Attribute):
    """
    Position attribute class to manage its properties in the ARENA: 3D object position.
    Usage: position=Position(x,y,z)

    :param float x:  (optional)
    :param float y:  (optional)
    :param float z:  (optional)
    """

    def __init__(self, x=None, y=None, z=None):
        if x is not None and (y is None or z is None or isinstance(x, Iterable)):
            raise ValueError("Position takes x,y,z")
        x = x or 0
        y = y or 0
        z = z or 0
        super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z))

    def to_str(self):
        return Utils.tuple_to_string((self.x, self.y, self.z))

    def distance_to(self, pos):
        return math.sqrt(
            (self.x - pos.x) ** 2 + (self.y - pos.y) ** 2 + (self.z - pos.z) ** 2
        )



from .arena_object import Object

class ArenauiCard(Object):
    """
    ArenauiCard object class to manage its properties in the ARENA: ARENAUI element which displays text and optionally an image.
    
    :param str body: This is the text body of the card. (optional)
    :param str bodyAlign: Body Text Alignment Allows [left, center, right, justify] Defaults to 'left' (optional)
    :param bool closeButton: Show close button (optional)
    :param str font: Font to use for button text. Allows [Roboto, Roboto-Mono] Defaults to 'Roboto' (optional)
    :param float fontSize: Font Size Defaults to '0.035' (optional)
    :param str img: This image will be embedded alongside the body text. (optional)
    :param str imgCaption: This will caption the image. (optional)
    :param str imgDirection: Image Direction Allows [left, right] Defaults to 'right' (optional)
    :param str imgSize: Image sizing Allows [cover, contain, stretch] Defaults to 'cover' (optional)
    :param float textImageRatio: Text to Image Width Ratio Defaults to '0.5' (optional)
    :param str theme: Color Theme Allows [light, dark] Defaults to 'light' (optional)
    :param str title: Title (optional)
    :param float widthScale: Width scale multiplier Defaults to '1' (optional)
    """
    object_type = "arenaui-card"

    def __init__(self, **kwargs):
        super().__init__(object_type=ArenauiCard.object_type, **kwargs)


class Card(ArenauiCard):
    """
    Alternate name for ArenauiCard.
    """



from ..utils import Utils
from .attribute import Attribute
from collections.abc import Iterable


class Scale(Attribute):
    """
    Scale attribute class to manage its properties in the ARENA: 3D object scale.
    Usage: scale=Scale(x,y,z)

    :param float x:  Defaults to '1' (optional)
    :param float y:  Defaults to '1' (optional)
    :param float z:  Defaults to '1' (optional)
    """

    def __init__(self, x=None, y=None, z=None):
        if x is not None and (y is None or z is None or isinstance(x, Iterable)):
            raise ValueError("Scale takes x,y,z")
        x = x or 1
        y = y or 1
        z = z or 1
        super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z))

    def to_str(self):
        return Utils.tuple_to_string((self.x, self.y, self.z))



from .arena_object import Object

class Tetrahedron(Object):
    """
    Tetrahedron object class to manage its properties in the ARENA: Tetrahedron Geometry.
    
    :param int detail: detail (optional)
    :param float radius: radius Defaults to '1' (optional)
    """
    object_type = "tetrahedron"

    def __init__(self, **kwargs):
        super().__init__(object_type=Tetrahedron.object_type, **kwargs)



from ..utils import Utils
from .attribute import Attribute
from .position import Position

class Landmark(Attribute):
    """
    Landmark attribute class to manage its properties in the ARENA: Define entities as a landmark; Landmarks appears in the landmark list and you can move (teleport) to them; You can define the behavior of the teleport: if you will be at a fixed or random distance, looking at the landmark, fixed offset or if it is constrained by a navmesh (when it exists).
    Usage: landmark=Landmark(...)
    
    :param str constrainToNavMesh: Teleports should snap to navmesh. Allows [false, any, coplanar] Defaults to 'false' (optional)
    :param str label: Landmark description to display in the landmark list. (optional)
    :param bool lookAtLandmark: Set to true to make users face the landmark when teleported to it. Defaults to 'True' (optional)
    :param dict offsetPosition: Use as a static teleport x,y,z offset. Defaults to '{'x': 0, 'y': 1.6, 'z': 0}' (optional)
    :param float randomRadiusMax: Maximum radius from the landmark to teleport to. (optional)
    :param float randomRadiusMin: Minimum radius from the landmark to teleport to. (randomRadiusMax must > 0). (optional)
    :param bool startingPosition: Set to true to use this landmark as a scene start (spawn) position. If several landmarks with startingPosition=true exist in a scene, one will be randomly selected. (optional)
    """
    def __init__(self, **kwargs):
        if "offsetPosition" in kwargs:
            if isinstance(kwargs["offsetPosition"], tuple) or isinstance(kwargs["offsetPosition"], list):
                kwargs["offsetPosition"] = Utils.tuple_to_string(kwargs["offsetPosition"])
            # can get away with using Position.to_str() even if animation is rotation (euler) or scale
            # since we just want the 3 values to be a space separated string
            elif isinstance(kwargs["offsetPosition"], Position):
                kwargs["offsetPosition"] = kwargs["offsetPosition"].to_str()
        super().__init__(**kwargs)



from .attribute import Attribute

class Sound(Attribute):
    """
    Sound attribute class to manage its properties in the ARENA: The sound component defines the entity as a source of sound or audio. The sound component is positional and is thus affected by the component's position.  More properties at (https://aframe.io/docs/1.5.0/components/sound.html) A-Frame Sound.
    Usage: sound=Sound(...)
    
    :param bool autoplay: Whether to automatically play sound once set. (optional)
    :param str distanceModel: Sound model. Allows [linear, inverse, exponential] Defaults to 'inverse' (optional)
    :param bool loop: Whether to loop the sound once the sound finishes playing. (optional)
    :param float maxDistance: Maximum distance between the audio source and the listener, after which the volume is not reduced any further. Defaults to '10000' (optional)
    :param str on: An event for the entity to listen to before playing sound. Allows [mousedown, mouseup, mouseenter, mouseleave, triggerdown, triggerup, gripdown, gripup, menudown, menuup, systemdown, systemup, trackpaddown, trackpadup] Defaults to 'mousedown' (optional)
    :param float poolSize: Numbers of simultaneous instances of this sound that can be playing at the same time. Defaults to '1' (optional)
    :param bool positional: Whether or not the audio is positional (movable). Defaults to 'True' (optional)
    :param float refDistance: Reference distance for reducing volume as the audio source moves further from the listener. Defaults to '1' (optional)
    :param float rolloffFactor: Describes how quickly the volume is reduced as the source moves away from the listener. Defaults to '1' (optional)
    :param str src: URL path to sound file e.g. 'store/users/wiselab/sound/wave.mp3'. (optional)
    :param float volume: How loud to play the sound. Defaults to '1' (optional)
    """
    def __init__(self, src, positional=False, **kwargs):
        super().__init__(src=src, positional=positional, **kwargs)



from .arena_object import Object

class Ring(Object):
    """
    Ring object class to manage its properties in the ARENA: Ring Geometry.
    
    :param float radiusInner: radius inner Defaults to '0.8' (optional)
    :param float radiusOuter: radius outer Defaults to '1.2' (optional)
    :param int segmentsPhi: segments phi Defaults to '10' (optional)
    :param int segmentsTheta: segments theta Defaults to '32' (optional)
    :param float thetaLength: theta length Defaults to '360' (optional)
    :param float thetaStart: theta start (optional)
    """
    object_type = "ring"

    def __init__(self, **kwargs):
        super().__init__(object_type=Ring.object_type, **kwargs)



from .attribute import Attribute


class RemoteRender(Attribute):
    """
    RemoteRender attribute class to manage its properties in the ARENA: Whether or not an object should be remote rendered [Experimental].
    Usage: remote_render=RemoteRender(...)
    
    :param bool enabled: Remote Render this object. Defaults to 'True' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .arena_object import Object

class Videosphere(Object):
    """
    Videosphere object class to manage its properties in the ARENA: Videosphere Geometry.
    
    :param bool autoplay: Autoplay Defaults to 'True' (optional)
    :param str crossOrigin: Cross Origin Defaults to 'anonymous' (optional)
    :param bool loop: Loop Defaults to 'True' (optional)
    :param float radius: Radius Defaults to '5000' (optional)
    :param int segmentsHeight: Segments Height Defaults to '64' (optional)
    :param int segmentsWidth: Segments Width Defaults to '64' (optional)
    :param str src: URI, relative or full path of an image/video file. e.g. 'store/users/wiselab/images/360falls.mp4'. (optional)
    """
    object_type = "videosphere"

    def __init__(self, **kwargs):
        super().__init__(object_type=Videosphere.object_type, **kwargs)



from .arena_object import Object

class Plane(Object):
    """
    Plane object class to manage its properties in the ARENA: Plane Geometry.
    
    :param float height: height Defaults to '1' (optional)
    :param int segmentsHeight: segments height Defaults to '1' (optional)
    :param int segmentsWidth: segments width Defaults to '1' (optional)
    :param float width: width Defaults to '1' (optional)
    """
    object_type = "plane"

    def __init__(self, **kwargs):
        super().__init__(object_type=Plane.object_type, **kwargs)



from .attribute import Attribute

class GotoUrl(Attribute):
    """
    GotoUrl attribute class to manage its properties in the ARENA: Goto given URL; Requires click-listener.
    Usage: goto_url=GotoUrl(...)
    
    :param str dest:  Allows [popup, newtab, sametab] Defaults to 'sametab' (optional)
    :param str on:  Allows [mousedown, mouseup] Defaults to 'mousedown' (optional)
    :param str url: Landmark Url (optional)
    """
    def __init__(self, url, dest="static", on="mousedown", **kwargs):
        if dest != "popup" and dest != "newtab" and dest != "sametab":
            dest = "popup"
        super().__init__(url=url, dest=dest, on=on, **kwargs)



from ..base_object import *
from .arena_object import Object
from ..attributes import Data
import uuid
import os

from ..env import (
    PROGRAM_OBJECT_ID,
)

class Program(BaseObject):
    """
    Program object in an ARENA Scene.

    :param str name: Name of the program 
    :param str affinity: Indicates the module affinity (client=client's runtime; none or empty=any suitable/available runtime)
    :param str instantiate: Single instance of the program (=single), or let every client create a program instance (=client). Per client instance will create new uuid for each program.
    :param str filename: Filename of the entry binary
    :param str filetype: Type of the program (WA=WASM or PY=Python)
    :param str parent: Request to deploy to this runtime (can be a runtime name or UUID); usually left blank.
    :param str[] args: Command-line arguments (passed in argv); e.g. [ "arg=value" ]. 
    :param str[] env: Environment variables; e.g. [ "SCENE=ascene" ].
    
    """

    type = "program"
    object_type = "program"
 
    def __init__(self, object_id=str(uuid.uuid4()), persist=False, objects_list_add=True, **kwargs):
                            
        # remove timestamp, if exists
        if "timestamp" in kwargs: del kwargs["timestamp"]

        # remove "updatedAt", if exists
        if "updatedAt" in kwargs: del kwargs["updatedAt"]

        # remove "action", if exists
        if "action" in kwargs: del kwargs["action"]
                    
        # update program object or print warning if existing object
        if Object.exists(object_id):
            if not Object.get(object_id).persist:
                print("[WARNING]", f"An object with object_id of {object_id} was already created. The previous object will be overwritten.")
            Object.remove(Object.get(object_id))

        # set default persist to False
        self.persist = False
        
        # setup attributes in the "data" field
        data = kwargs.get("data", kwargs)
        data = Data(**data)
        super().__init__(
                object_id=object_id,
                type=Program.type,
                persist=persist,
                data=data
            )

        # add current object to all_objects dict
        if objects_list_add: Object.add(self)

    def update_attributes(self, evt_handler=None, update_handler=None, **kwargs):
        if "data" not in self:
            return

        # update "persist"
        self.persist = kwargs.get("persist", self.persist)

        data = self.data
        Data.update_data(data, kwargs)

    def json(self, **kwargs):
        json_data = {}
        json_payload = { k: v for k, v in vars(self).items() }
        json_payload.update(kwargs)
        data = vars(json_payload["data"])

        for k,v in data.items():
            if v is None:
                json_data[k] = v

        json_payload["data"] = data
        return self.json_encode(json_payload)



import asyncio
import json
import os
import re
import sys
from datetime import datetime
from inspect import signature
import threading
import uuid

from .arena_mqtt import ArenaMQTT
from .attributes import *
from .events import *
from .objects import *
from .utils import Utils, ArenaTelemetry, ArenaCmdInterpreter, ProgramRunInfo, QueueStats

from .env import (
    SCENE,
    PROGRAM_OBJECT_ID,
    _get_env
)

class Scene(ArenaMQTT):
    """
    Gives access to an ARENA scene.
    Can create and execute various user-defined functions/tasks.

    :param str host: Hostname of the ARENA webserver (required).
    :param str realm: Reserved topic fork for future use (optional).
    :param str namespace: Username of authenticated user or other namespace (automatic).
    :param str scene: The name of the scene, without namespace (required).
    """

    def __init__(
                self,
                host = "arenaxr.org",
                realm = "realm",
                network_latency_interval = 10000,  # run network latency update every 10s
                on_msg_callback = None,
                new_obj_callback = None,
                user_join_callback = None,
                user_left_callback = None,
                delete_obj_callback = None,
                end_program_callback = None,
                video = False,
                debug = False,
                cli_args = False,
                **kwargs
            ):
                
        # init telemetry
        self.telemetry = ArenaTelemetry()        
        
        # setup event to let others wait on connection
        self.connected_evt = threading.Event()
                
        # start the command interpreter (if enabled by env variable)
        self.cmd_interpreter = ArenaCmdInterpreter(self, show_attrs=('config_data', 'scene', 'users', 'all_objects', 'run_info'),
                                    get_callables=('persisted_objs', 'persisted_scene_option', 'writable_scenes', 'user_list'),
                                    start_cmd_event=self.connected_evt)
                
        if cli_args:
            self.args = self.parse_cli()
            if self.args["host"]:
                kwargs["host"] = self.args["host"]
            if self.args["namespace"]:
                kwargs["namespace"] = self.args["namespace"]
            if self.args["scene"]:
                kwargs["scene"] = self.args["scene"]
            if self.args["debug"]:
                debug = self.args["debug"]

        if os.environ.get(SCENE):
            self.scene = _get_env(SCENE)
            print(f"Using Scene from 'SCENE' env variable: {self.scene}")
        elif "scene" in kwargs and kwargs["scene"]:
            self.scene = kwargs["scene"]
            print(f"Using Scene from 'scene' input parameter: {self.scene}")
        else:
            self.exit("Scene argument (scene) is unspecified or None, aborting...")

        if re.search("/", self.scene):
            self.exit("Scene cannot include '/', aborting...")
            
        super().__init__(
            host,
            realm,
            network_latency_interval,
            on_msg_callback,
            end_program_callback,
            video,
            debug,
            **kwargs
        )

        with self.telemetry.start_span(f"init {self.namespace}/{self.scene}") as span:
            # 'init' span will track the remainder of the initialization

            # create a program object to describe this program
            # PROGRAM_OBJECT_ID allows to match the object id of persisted program object
            # when a program object with PROGRAM_OBJECT_ID is loaded from persist, it will replace this one
            self.program = Program(object_id=_get_env(PROGRAM_OBJECT_ID, super().client_id()), 
                                   name=f"{self.namespace}/{self.scene}", 
                                   filename=sys.argv[0], 
                                   filetype="PY")
                        
            self.persist_host = self.config_data["ARENADefaults"]["persistHost"]
            self.persist_path = self.config_data["ARENADefaults"]["persistPath"]

            self.persist_url = f"https://{self.persist_host}{self.persist_path}{self.namespaced_target}"

            # set up callbacks
            self.new_obj_callback = new_obj_callback
            self.delete_obj_callback = delete_obj_callback
            self.user_join_callback = user_join_callback
            self.user_left_callback = user_left_callback

            self.unspecified_object_ids = set() # objects that exist in the scene,
                                                # but this scene instance does not
                                                # have a reference to
            self.users = {} # dict of all users

            # setup program run info to collect stats
            self.run_info = ProgramRunInfo(self.event_loop,
                                           queue_len_callable=lambda: self.get_rcv_pub_queue_len(), 
                                           update_callback=self.run_info_update, 
                                           web_host=self.web_host,
                                           namespace=self.namespace,
                                           scene=self.scene,
                                           realm=self.realm
                                           )

            # Always use the the hostname specified by the user, or defaults.
            print(f"Loading: https://{self.web_host}/{self.namespace}/{self.scene}, realm={self.realm}")
                        
            span.add_event(f"Loading: https://{self.web_host}/{self.namespace}/{self.scene}, realm={self.realm}")

    def exit(self, arg=0):
        """Custom exit to push errors to telemetry"""
        error_msg=None
        if arg != 0:
            error_msg = f"Exiting with sys.exit('{arg}')"
        self.telemetry.exit(error_msg)
        os._exit(arg)
            
    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        if rc == 0:
            # set event
            self.connected_evt.set()

            # create arena-py Objects from persist server
            # no need to return anything here            
            self.get_persisted_objs()

    def on_message(self, client, userdata, msg):
        super().on_message(client, userdata, msg)
        self.run_info.msg_rcv()

    def on_publish(self, client, userdata, mid):
        self.run_info.msg_publish()
        
    async def process_message(self):
        while True:
            try:
                msg = await self.msg_queue.get()
            except RuntimeError as e:
                self.telemetry.add_event(f"Ignoring error: {e}")                
                return

            # extract payload
            try:
                payload_str = msg.payload.decode("utf-8", "ignore")
                payload = json.loads(payload_str)
            except Exception as e:
                self.telemetry.add_event(f"Malformed payload: {payload_str}. {e}.")
                return

            object_id = payload.get("object_id", None)
            action = payload.get("action", None)

            with self.telemetry.start_process_msg_span(object_id, action) as span:    
                try:
                    # update object attributes, if possible
                    if object_id:
                        data = payload.get("data", {})
                        object_type = data.get("object_type", None)

                        event = None

                        # create/get object from object_id
                        if object_id in self.all_objects:
                            obj = self.all_objects[object_id]
                        else:
                            ObjClass = OBJECT_TYPE_MAP.get(object_type, Object)
                            obj = ObjClass(**payload)

                        # react to action accordingly
                        if action:
                            if action == "clientEvent":
                                event = Event(**payload)
                                if obj.evt_handler:
                                    self.callback_wrapper(obj.evt_handler, event, payload)
                                    continue
                                span.add_event("Client event: {event}")


                            elif action == "delete":
                                if Camera.object_type in object_id: # object is a camera
                                    if object_id in self.users:
                                        if self.user_left_callback:
                                            self.callback_wrapper(
                                                    self.user_left_callback,
                                                    self.users[object_id],
                                                    payload
                                                )
                                        del self.users[object_id]
                                elif HandLeft.object_type in object_id or HandRight.object_type in object_id: # object is a hand/controller
                                    if "dep" in obj.data:
                                        user_id = obj.data.dep
                                        if user_id in self.users:
                                            user = self.users[user_id]
                                            if obj in user.hands.values():
                                                if user.hand_remove_callback:
                                                    self.callback_wrapper(
                                                            user.hand_remove_callback,
                                                            obj,
                                                            payload
                                                        )
                                                hand_key = HandLeft.object_type if HandLeft.object_type in object_id else HandRight.object_type
                                                del user.hands[hand_key]
                                elif self.delete_obj_callback:
                                    self.callback_wrapper(self.delete_obj_callback, obj, payload)
                                Object.remove(obj)
                                span.add_event("Object delete.")

                                continue

                            else: # create/update
                                obj.update_attributes(**payload)
                                span.add_event("Object attributes update.")
                        
                        else:
                            self.telemetry.set_error("No message action!", span)

                        # call new message callback for all messages
                        if self.on_msg_callback:
                            if not event:
                                self.callback_wrapper(self.on_msg_callback, obj, payload)
                            else:
                                self.callback_wrapper(self.on_msg_callback, event, payload)

                        if object_type:
                            # run user_join_callback when user is found
                            if object_type == Camera.object_type:
                                if object_id not in self.users:
                                    if object_id in self.all_objects:
                                        self.users[object_id] = obj
                                    else:
                                        self.users[object_id] = Camera(**payload)

                                    if self.user_join_callback:
                                        self.callback_wrapper(
                                                self.user_join_callback,
                                                self.users[object_id],
                                                payload
                                            )

                            elif object_type == HandLeft.object_type or object_type == HandRight.object_type:
                                if "dep" in obj.data:
                                    user_id = obj.data.dep
                                    if user_id in self.users:
                                        user = self.users[user_id]
                                        if obj not in user.hands.values():
                                            user.hands[object_type] = obj
                                            obj.camera = user

                                            if user.hand_found_callback:
                                                self.callback_wrapper(
                                                    user.hand_found_callback,
                                                    obj,
                                                    payload
                                                )

                        # if its an object the library has not seen before, call new object callback
                        elif object_id not in self.unspecified_object_ids and self.new_obj_callback:
                            self.callback_wrapper(self.new_obj_callback, obj, payload)
                            self.unspecified_object_ids.add(object_id)
                            span.add_event("New Object.")
                        
                        span.add_event("Handle Msg Done.")
                    else:
                        self.telemetry.set_error("No object id!", span)

                except Exception as e:                    
                    self.telemetry.set_error(f"Something went wrong, ignoring: {payload}. {e}")

    def callback_wrapper(self, func, arg, msg):
        """Checks for number of arguments for callback"""
        if len(signature(func).parameters) != 3:
            print("[DEPRECATED]", "Callbacks and handlers now take 3 arguments: (scene, obj/evt, msg)!")
            func(arg)
        else:
            func(self, arg, msg)

    def generate_custom_event(self, evt, action="clientEvent"):
        """Publishes an custom event. Could be user or library defined"""
        return self._publish(evt, action)

    def generate_click_event(self, obj, type="mousedown", **kwargs):
        """Publishes an click event"""
        _type = type
        evt = Event(object_id=obj.object_id,
                    type=_type,
                    position=obj.data.position,
                    source=self.mqttc_id,
                    **kwargs)
        return self.generate_custom_event(evt, action="clientEvent")

    def manipulate_camera(self, cam, **kwargs):
        """Publishes a camera manipulation event"""
        if kwargs.get("position"):
            if isinstance(kwargs["position"], (list, tuple)):
                kwargs["position"] = Position(*kwargs["position"])
            elif isinstance(kwargs["position"], dict):
                kwargs["position"] = Position(**kwargs["position"])

        if kwargs.get("rotation"):
            if isinstance(kwargs["rotation"], (list, tuple)):
                kwargs["rotation"] = Rotation(*kwargs["rotation"])
            elif isinstance(kwargs["rotation"], dict):
                kwargs["rotation"] = Rotation(**kwargs["rotation"])

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type=Camera.object_type,
                    **kwargs)
        return self.generate_custom_event(evt, action="update")

    def look_at(self, cam, target):
        """Publishes a camera manipulation event"""
        if isinstance(target, tuple) or isinstance(target, list):
            target = Position(*target)
        elif isinstance(target, dict):
            target = Position(**target)
        elif isinstance(target, Object):
            target = target.object_id

        if isinstance(cam, Object):
            object_id = cam.object_id
        else:
            object_id = cam
        evt = Event(object_id=object_id,
                    type="camera-override",
                    object_type="look-at",
                    target=target)
        return self.generate_custom_event(evt, action="update")

    @property
    def all_objects(self):
        """Returns all the objects in a scene"""
        return Object.all_objects

    def add_object(self, obj):
        """Public function to create an object"""
        res = self._publish(obj, "create")
        self.run_animations(obj)
        return res

    def add_objects(self, objs):
        """Public function to create multiple objects in a list"""
        for obj in objs:
            self.add_object(obj)
        return len(objs)

    def update_object(self, obj, **kwargs):
        """Public function to update an object"""
        if kwargs:
            obj.update_attributes(**kwargs)

        # Check if any keys in delayed_prop_tasks are pending new animations
        # and cancel corresponding final update tasks or, if they are in
        # kwarg property updates, cancel the task as well as the animation
        need_to_run_animations = False
        if len(obj.delayed_prop_tasks) > 0:
            for anim in obj.animations:
                if anim.property in obj.delayed_prop_tasks:
                    task_to_cancel = obj.delayed_prop_tasks[anim.property]
                    task_to_cancel.cancel()
                    del task_to_cancel

            for k in kwargs:
                if str(k) in obj.delayed_prop_tasks:
                    need_to_run_animations = True
                    task_to_cancel = obj.delayed_prop_tasks[k]
                    task_to_cancel.cancel()
                    del task_to_cancel
                    obj.dispatch_animation(
                        Animation(property=k, enabled=False, dur=0)
                    )
            if need_to_run_animations:
                self.run_animations(obj)

        res = self._publish(obj, "update")
        self.run_animations(obj)
        return res

    def update_objects(self, objs, **kwargs):
        """Public function to update multiple objects in a list"""
        for obj in objs:
            self.update_object(obj, **kwargs)
        return len(objs)

    def delete_object(self, obj):
        """Public function to delete an object"""
        payload = {
            "object_id": obj.object_id
        }
        Object.remove(obj)
        return self._publish(payload, "delete", custom_payload=True)

    def delete_attributes(self, obj, attributes=None):
        """Public function to delete a list of 'attributes' as a string[], updating each to null"""
        updated_data = {}
        for attr in attributes:
            obj.data[attr] = None  # remove from large internal storage
            updated_data[attr] = None # remove from small external publish
        payload = {
            "object_id": obj.object_id,
            "type": obj.type,
            "persist": obj.persist,
            "data": updated_data  # dashes handled from string array
        }
        return self._publish(payload, "update", custom_payload=True)

    def run_animations(self, obj):
        """Runs all dispatched animations"""
        if obj.animations:
            payload = {
                "object_id": obj.object_id,
                "type": obj.type,
                "data": {"object_type": obj.data.object_type}
            }
            for i,animation in enumerate(obj.animations):
                if isinstance(animation, AnimationMixer):
                    payload["data"][f"animation-mixer"] = vars(animation)
                else:
                    anim = vars(animation).copy()
                    payload["data"][f"animation__{anim['property']}"] = anim
                    Utils.dict_key_replace(anim, "start", "from")
                    Utils.dict_key_replace(anim, "end", "to")
                    self.create_delayed_task(obj, anim)
            obj.clear_animations()
            return self._publish(payload, "update", custom_payload=True)

    def create_delayed_task(self, obj, anim):
        """
        Creates a delayed task to push the end state of an animation after the expected
        duration. Uses async sleep to avoid blocking.
        :param obj: arena object to update
        :param anim: Animation to run
        :return: created async task
        """

        async def _delayed_task():
            try:
                sleep_dur = (anim.get('dur', 0) + anim.get('delay', 0)) / 1000
                await asyncio.sleep(sleep_dur)  # this is in seconds
                final_state = anim["from"] if anim.get("dir", "normal") == "reverse"\
                    else anim["to"]
                obj.update_attributes(**{anim["property"]: final_state})
                self.update_object(obj)
                obj.delayed_prop_tasks.pop(anim["property"], None)
            except asyncio.CancelledError:
                self.telemetry.add_event(f"Animation end task cancelled for {obj.object_id}.{anim['property']}")

        if anim.get("loop", 0) or anim.get("enabled", True) is False:
            return None

        delayed_task = asyncio.create_task(_delayed_task())
        obj.delayed_prop_tasks[anim["property"]] = delayed_task
        delayed_task.object_id = obj.object_id
        return delayed_task

    def _publish(self, obj, action, custom_payload=False):
        """Publishes to mqtt broker with "action":action"""
        with self.telemetry.start_publish_span(obj.object_id, action, obj.type) as span:
            if not self.can_publish:
                self.telemetry.set_error(f"ERROR: Publish failed! You do not have permission to publish to topic {self.root_topic} on {self.web_host}", span)

            topic = f"{self.root_topic}/{self.mqttc_id}/{obj['object_id']}"
            d = datetime.utcnow().isoformat()[:-3]+"Z"

            if custom_payload:
                payload = obj
                payload["action"] = action
                payload["timestamp"] = d
                payload = json.dumps(payload)
            else:
                payload = obj.json(action=action, timestamp=d)

            self.mqttc.publish(topic, payload, qos=0)
            if self.debug: 
                self.telemetry.add_event(f"[publish] {topic} {payload}")
            return payload

    def get_persisted_obj(self, object_id):
        """Returns a dictionary for a persisted object.
        
           If object is known by arena-py, return local object, not persisted
        """
        obj = None
        if object_id in self.all_objects:
            obj = self.all_objects[object_id]
            obj.persist = True
        else:
            # pass token to persist
            data = self.auth.urlopen(url=f"{self.persist_url}/{object_id}", creds=True)
            persist_obj = json.loads(data)
            if len(obj) > 0:
                obj = obj[0]

                obj_id = persist_obj.get("object_id", object_id)
                data = persist_obj.get("attributes", {})
                object_type = data.get("object_type")

                # special case for Program type
                if persist_obj.get("type") == Program.object_type: object_type = Program.object_type

                if object_type != None:
                    obj_class = OBJECT_TYPE_MAP.get(object_type, Object)
                    obj = obj_class(object_id=obj_id, data=data)
                    obj.persist = True
                    
        return obj
    
    def get_persisted_objs(self):
        """Returns a dictionary of persisted objects.
        
           If object is known by arena-py, return our local object, not persisted
           Silently fails/skip objects without object_id and object_type (except programs)
           Instanciates generic Object if object_type is given but unknown to arena-py
        """
        objs = {}
        # pass token to persist
        data = self.auth.urlopen(url=self.persist_url, creds=True)
        output = json.loads(data)
        for obj in output:
            # note: no exception on missing fields
            object_id = obj.get("object_id")
            data = obj.get("attributes", {})
            object_type = data.get("object_type")
            
            # special case for Program type
            if obj.get("type") == Program.object_type: 
                object_type = Program.object_type
            
            if object_id != None:
                if object_id in self.all_objects:
                    # note: object from our list (not from persist)
                    persisted_obj = self.all_objects[object_id]
                    persisted_obj.persist = True
                elif object_type != None:
                    # note: instanciate even with empty attributes if object_type is unknown to arena-py
                    obj_class = OBJECT_TYPE_MAP.get(object_type, Object)
                    persisted_obj = obj_class(object_id=object_id, data=data)
                    persisted_obj.persist = True
                    
                    # replace program object, if matches our program id
                    if object_type == Program.object_type:
                        if os.environ.get(PROGRAM_OBJECT_ID) == object_id:
                            persisted_obj.persist = True
                            self.program = persisted_obj
                        else:
                            # dont persist program objet if not ours
                            persisted_obj.persist = False
                            
            
                objs[object_id] = persisted_obj

        return objs

    def get_persisted_scene_option(self):
        """Returns a dictionary for scene-options. [TODO] wrap the output as a BaseObject"""
        scene_opts_url = f"{self.persist_url}?type=scene-options"
        # pass token to persist
        data = self.auth.urlopen(url=scene_opts_url, creds=True )
        output = json.loads(data)
        return output

    def get_writable_scenes(self):
        """ Request list of scene names for logged in user account that user has publish permission for.
        Returns: list of scenes.
        """
        return self.auth.get_writable_scenes(web_host=self.web_host)

    def get_user_list(self):
        """Returns a list of users"""
        return self.users.values()

    def get_rcv_pub_queue_len(self):
        """Return QueueStats object with receive and publish queue length"""
        return QueueStats(super().rcv_queue_len(),  super().pub_queue_len())
    
    def run_info_update(self, run_info):
        """Callbak when program stats are updated; publish program object update"""
        # Add run info to program data object and publish program object update
        run_info.add_program_info(self.program.data)
        self._publish(self.program, "update")
        
class Arena(Scene):
    """
    Another name for Scene.
    """



from .arena_object import Object

class PcdModel(Object):
    """
    PcdModel object class to manage its properties in the ARENA: Load a PCD model.  Format: (https://pointclouds.org/documentation/tutorials/index.html) Point Clouds. See guidance to store paths under (https://docs.arenaxr.org/content/interface/filestore.html) ARENA File Store, CDN, or DropBox.
    
    :param str pointColor: Color of the points. Defaults to '#7f7f7f' (optional)
    :param float pointSize: Size of the points. Defaults to '0.01' (optional)
    :param str url: Use File Store paths under 'store/users/username', see CDN and other storage options in the description above. (optional)
    """
    object_type = "pcd-model"

    def __init__(self, **kwargs):
        super().__init__(object_type=PcdModel.object_type, **kwargs)



from .arena_object import Object


class Ocean(Object):
    """
    Ocean object class to manage its properties in the ARENA: Flat-shaded ocean primitive.

    :param float amplitude: Wave amplitude. Defaults to '0.1' (optional)
    :param float amplitudeVariance: Wave amplitude variance. Defaults to '0.3' (optional)
    :param str color: Wave color. Defaults to '#7AD2F7' (optional)
    :param float density: Density of waves. Defaults to '10' (optional)
    :param float depth: Depth of the ocean area. Defaults to '10' (optional)
    :param float opacity: Wave opacity. Defaults to '0.8' (optional)
    :param float speed: Wave speed. Defaults to '1' (optional)
    :param float speedVariance: Wave speed variance. Defaults to '2' (optional)
    :param float width: Width of the ocean area. Defaults to '10' (optional)
    """
    object_type = "ocean"

    def __init__(self, width=10, depth=10, **kwargs):
        # NOTE: The 'ocean' component for a-frame requires at least depth or width before it will load.
        super().__init__(object_type=Ocean.object_type, width=width, depth=depth, **kwargs)



from .arena_object import Object
from ..attributes import Position

class Line(Object):
    """
    Line object class to manage its properties in the ARENA: Draw a line.
    
    :param str color: Line color. Defaults to '#74BEC1' (optional)
    :param dict end: End coordinate. Defaults to '{'x': -0.5, 'y': -0.5, 'z': 0}' (optional)
    :param float opacity: Line opacity. Defaults to '1' (optional)
    :param dict start: Start point coordinate. Defaults to '{'x': 0, 'y': 0.5, 'z': 0}' (optional)
    :param bool visible: Whether the material is visible. Defaults to 'True' (optional)
    """
    object_type = "line"

    def __init__(self, start=(0,0,0), end=(10,10,10), **kwargs):
        super().__init__(object_type=Line.object_type, start=start, end=end, **kwargs)



from ..utils import Utils
from .attribute import Attribute
from .position import Position
from .rotation import Rotation
from .scale import Scale

class Animation(Attribute):
    """
    Animation attribute class to manage its properties in the ARENA: Animate and tween values.  More properties at (https://aframe.io/docs/1.5.0/components/animation.html) A-Frame Animation component. Easing properties are detailed at (https://easings.net) easings.net.
    Usage: animation=Animation(...)
    
    :param bool autoplay: Whether or not the animation should autoplay. Should be specified if the animation is defined for the animation-timeline component (currently not supported). Defaults to 'True' (optional)
    :param float delay: How long (milliseconds) to wait before starting. (optional)
    :param str dir: Which dir to go from from to to. Allows [normal, alternate, reverse] Defaults to 'normal' (optional)
    :param float dur: How long (milliseconds) each cycle of the animation is. Defaults to '1000' (optional)
    :param str easing: Easing function of animation. To ease in, ease out, ease in and out. See easings.net for more. Allows [easeInQuad, easeInCubic, easeInQuart, easeInQuint, easeInSine, easeInExpo, easeInCirc, easeInBack, easeInElastic, easeOutQuad, easeOutCubic, easeOutQuart, easeOutQuint, easeOutSine, easeOutExpo, easeOutCirc, easeOutBack, easeOutElastic, easeInOutQuad, easeInOutCubic, easeInOutQuart, easeInOutQuint, easeInOutSine, easeInOutExpo, easeInOutCirc, easeInOutBack, easeInOutElastic, linear] Defaults to 'easeInQuad' (optional)
    :param float elasticity: How much to bounce (higher is stronger). Defaults to '400' (optional)
    :param bool enabled: If disabled, animation will stop and startEvents will not trigger animation start. Defaults to 'True' (optional)
    :param str from: Initial value at start of animation. If not specified, the current property value of the entity will be used (will be sampled on each animation start). It is best to specify a from value when possible for stability. (optional)
    :param bool isRawProperty: Flag to animate an arbitrary object property outside of A-Frame components for better performance. If set to true, for example, we can set property to like components.material.material.opacity. If property starts with components or object3D, this will be inferred to true. (optional)
    :param str loop: How many times the animation should repeat. If the value is true, the animation will repeat infinitely. (optional)
    :param list[dict] pauseEvents: Comma-separated list of events to listen to trigger pause. Can be resumed with resumeEvents. (optional)
    :param str property: Property to animate. Can be a component name, a dot-delimited property of a component (e.g., material.color), or a plain attribute. (optional)
    :param list[dict] resumeEvents: Comma-separated list of events to listen to trigger resume after pausing. (optional)
    :param bool round: Whether to round values. (optional)
    :param list[dict] startEvents: Comma-separated list of events to listen to trigger a restart and play. Animation will not autoplay if specified. startEvents will restart the animation, use pauseEvents to resume it. If there are other animation components on the entity animating the same property, those animations will be automatically paused to not conflict. (optional)
    :param str to: Target value at end of animation. (optional)
    :param str type: Right now only supports color for tweening isRawProperty color XYZ/RGB vector values. (optional)
    """
    def __init__(self, **kwargs):
        if "start" in kwargs:
            if isinstance(kwargs["start"], tuple) or isinstance(kwargs["start"], list):
                kwargs["start"] = Utils.tuple_to_string(kwargs["start"])
            elif isinstance(kwargs["start"], (Position, Rotation, Scale)):
                kwargs["start"] = vars(kwargs["start"])
        if "end" in kwargs:
            if isinstance(kwargs["end"], tuple) or isinstance(kwargs["end"], list):
                kwargs["end"] = Utils.tuple_to_string(kwargs["end"])
            elif isinstance(kwargs["end"], (Position, Rotation, Scale)):
                kwargs["end"] = vars(kwargs["end"])
        super().__init__(**kwargs)



# This file is auto-generated from github.com/arenaxr/arena-schema, changes here may be overwritten.

from .translate import *
from .attribute import *
from .color import *
from .data import *
from .animation import *
from .animation_mixer import *
from .armarker import *
from .attribution import *
from .blip import *
from .box_collision_listener import *
from .click_listener import *
from .dynamic_body import *
from .gltf_model_lod import *
from .gltf_morph import *
from .goto_landmark import *
from .goto_url import *
from .impulse import *
from .jitsi_video import *
from .landmark import *
from .material import *
from .material_extras import *
from .multisrc import *
from .position import *
from .remote_render import *
from .rotation import *
from .scale import *
from .shadow import *
from .sound import *
from .spe_particles import *
from .static_body import *
from .textinput import *
from .video_control import *

ATTRIBUTE_TYPE_MAP = {
    "attribute": Attribute,
    "color": Color,
    "data": Data,
    "animation": Animation,
    "animation_mixer": AnimationMixer,
    "armarker": Armarker,
    "attribution": Attribution,
    "blip": Blip,
    "box_collision_listener": BoxCollisionListener,
    "click_listener": ClickListener,
    "dynamic_body": DynamicBody,
    "gltf_model_lod": GltfModelLod,
    "gltf_morph": GltfMorph,
    "goto_landmark": GotoLandmark,
    "goto_url": GotoUrl,
    "impulse": Impulse,
    "jitsi_video": JitsiVideo,
    "landmark": Landmark,
    "material": Material,
    "material_extras": MaterialExtras,
    "multisrc": Multisrc,
    "position": Position,
    "remote_render": RemoteRender,
    "rotation": Rotation,
    "scale": Scale,
    "shadow": Shadow,
    "sound": Sound,
    "spe_particles": SpeParticles,
    "static_body": StaticBody,
    "textinput": Textinput,
    "video_control": VideoControl,
}



from ..utils import Utils
from .arena_object import Object
from ..attributes import Attribute, Position

class Thickline(Object):
    """
    Thickline object class to manage its properties in the ARENA: Draw a line that can have a custom width.
    
    :param str color: Line color. Defaults to '#000000' (optional)
    :param float lineWidth: Width of line in px. Defaults to '1' (optional)
    :param str lineWidthStyler: Allows defining the line width as a function of relative position p along the path of the line. By default it is set to a constant 1. You may also choose one of the preset functions. Allows [default, grow, shrink, center-sharp, center-smooth, sine-wave] Defaults to 'default' (optional)
    :param str path: Comma-separated list of x y z coordinates of the line vertices. Defaults to '-2 -1 0, 0 20 0, 10 -1 10' (optional)
    """
    object_type = "thickline"

    def __init__(self, path=None, lineWidth=1, **kwargs):
        # path for thickline is a string, ie (1,2,3) -> "1 2 3"
        if path is None:
            path = [Position(0, 0, 0), Position(10, 10, 10), Position(10, -10, 10)]
        path_str = ""
        for p in path:
            if isinstance(p, Position):
                p = p.to_str()
            elif isinstance(p, Attribute):
                p = Position(**p.__dict__).to_str()
            elif isinstance(p, tuple) or isinstance(p, list):
                p = Utils.tuple_to_string(p)
            elif isinstance(p, dict):
                p = Position(**p).to_str()
            path_str += p + ","
        path_str = path_str.rstrip(",")
        super().__init__(object_type=ThickLine.object_type, path=path_str, lineWidth=lineWidth, **kwargs)


class ThickLine(Thickline):
    """
    Alternate name for Thickline.
    """



from .arena_object import Object
from ..attributes import Position, Rotation


class Camera(Object):
    """
    Camera object class to manage its properties in the ARENA: Camera is the pose and arena-user component data representing a user avatar.
    
    :param dict arena_user:  (optional)
    """

    object_type = "camera"

    def __init__(self, object_id, **kwargs):
        data = kwargs.get("data", {})
        arena_user = data.get("arena-user", {})

        self.hasAudio = arena_user.get("hasAudio", False)
        self.hasVideo = arena_user.get("hasVideo", False)
        self.hasAvatar = arena_user.get("hasAvatar", False)
        self.displayName = arena_user.get("displayName", "")
        self.jistsiId = arena_user.get("jistsiId", None)

        self.hands = {}
        self.hand_found_callback = None
        self.hand_remove_callback = None

        position = data.get("position", None)
        rotation = data.get("rotation", None)

        if position is not None and rotation is not None:
            super().__init__(
                object_type=Camera.object_type,
                object_id=object_id,
                position=Position(**position),
                rotation=Rotation(**rotation),
                **kwargs
            )
        elif position is not None:
            super().__init__(
                object_type=Camera.object_type,
                object_id=object_id,
                position=Position(**position),
                **kwargs
            )
        elif rotation is not None:
            super().__init__(
                object_type=Camera.object_type,
                object_id=object_id,
                rotation=Rotation(**rotation),
                **kwargs
            )



from .attribute import Attribute


class GltfMorph(Attribute):
    """
    GltfMorph attribute class to manage its properties in the ARENA: Allows you to target and control a gltf model's morphTargets created in Blender.  More properties at (https://github.com/elbobo/aframe-gltf-morph-component) A-Frame GLTF Morph component.
    Usage: gltf_morph=GltfMorph(...)

    :param str morphtarget: Name of morphTarget, can be found as part of the GLTF model. (optional)
    :param float value: Value that you want to set that morphTarget to (0 - 1). (optional)
    """

    def __init__(self, morphtarget, value):
        self.morphtarget = str(morphtarget)
        self.value = str(value)


class Morph(GltfMorph):
    """
    Alternate name for GltfMorph.
    Usage: gltf_morph=Morph(...)
    """



from .arena_object import Object

class TorusKnot(Object):
    """
    TorusKnot object class to manage its properties in the ARENA: Torus Knot Geometry.
    
    :param float p: P Defaults to '2' (optional)
    :param float q: Q Defaults to '3' (optional)
    :param float radius: radius Defaults to '1' (optional)
    :param float radiusTubular: radius tubular Defaults to '0.2' (optional)
    :param int segmentsRadial: segments radial Defaults to '8' (optional)
    :param int segmentsTubular: segments tubular Defaults to '100' (optional)
    """
    object_type = "torusKnot"

    def __init__(self, **kwargs):
        super().__init__(object_type=TorusKnot.object_type, **kwargs)



from .arena_object import Object

class Circle(Object):
    """
    Circle object class to manage its properties in the ARENA: Circle Geometry.
    
    :param float radius: radius Defaults to '1' (optional)
    :param int segments: segments Defaults to '32' (optional)
    :param float thetaLength: theta length Defaults to '360' (optional)
    :param float thetaStart: theta start (optional)
    """
    object_type = "circle"

    def __init__(self, **kwargs):
        super().__init__(object_type=Circle.object_type, **kwargs)



# Environment variables definitions
# When applicable, variable defaults are defined by ENV_DEFAULTS
#
import sys 
import os

REALM = "REALM"
"""
.. envvar:: REALM

The :envvar:`REALM` defines the ARENA Realm to listen to.
After connecting, the library listens to a scene topic as follows:
`{REALM}/s/{NAMESPACE}/{SCENE}`.
This variable overrides arguments passed in the scene's constructor.
"""

SCENE = "SCENE"
"""
.. envvar:: SCENE

The :envvar:`SCENE` defines ARENA Scene to listen to.
After connecting, the library listens to a scene topic as follows:
`{REALM}/s/{NAMESPACE}/{SCENE}`.
This variable overrides arguments passed in the scene's constructor.
"""

NAMESPACE = "NAMESPACE"
"""
.. envvar:: NAMESPACE

The :envvar:`NAMESPACE` defines ARENA Namespace to listen to.
After connecting, the library listens to a scene topic as follows:
`{REALM}/s/{NAMESPACE}/{SCENE}`.
This variable overrides arguments passed in the scene's constructor.
"""

ARENA_USERNAME = "ARENA_USERNAME"
"""
.. envvar:: ARENA_USERNAME

The :envvar:`ARENA_USERNAME` defines username used to authenticate.
If undefined, will try to use local authentication information previously saved.
"""

ARENA_PASSWORD = "ARENA_PASSWORD"
"""
.. envvar:: ARENA_PASSWORD

The :envvar:`ARENA_PASSWORD` defines password used to authenticate.
If undefined, will try to use local authentication information previously saved.
"""

MQTTH = "MQTTH"
"""
.. envvar:: MQTTH

The :envvar:`MQTTH` defines the MQTT host used by the library.
This variable allows to use a broker different from the host argument passed to the 
scene constructor
"""

DEVICE = "DEVICE"
"""
.. envvar:: DEVICE

The :envvar:`DEVICE` defines the name of a device, to publish and listen to. 
After connecting, the library listens to device topic as follows:
`{REALM}/d/{NAMESPACE}/{SCENE}`.

This variable overrides arguments passed in the command line.
"""

PROGRAM_OBJECT_ID = "PROGRAM_OBJECT_ID"
"""
.. envvar:: PROGRAM_OBJECT_ID

The :envvar:`PROGRAM_OBJECT_ID` indicates the object id in ARENA persist for this program.
This is passed by the runtime and used to identify the program object that represents the currently running program.
"""

ENABLE_INTERPRETER = "ENABLE_INTERPRETER"
"""
.. envvar:: ENABLE_INTERPRETER

The :envvar:`ENABLE_INTERPRETER` enables the a simple command line interpreter that
can be used to inspect library/program state. Set this variable with a value of 
`true`, `1` or `t` (case insensitive) to enable the interpreter.

Default: 'false'
"""

ARENA_TELEMETRY = "ARENA_TELEMETRY"
"""
.. envvar:: ARENA_TELEMETRY

The :envvar:`ARENA_TELEMETRY` environment variable enables the library's telemetry to generate 
traces, metrics, and logs. Set this variable with a value of `otlp`, `mqtt` or `console` (case insensitive) 
to enable telemetry using OpenTelemetry (OTEL) and its Protocol (OTLP), send JSON OTEL spans to MQTT, or to the console.
"""

OTLP_ENDPOINT = "OTLP_ENDPOINT"
"""
.. envvar:: OTLP_ENDPOINT

The :envvar:`OTLP_ENDPOINT` environment variable is used when OTLP telemetry is enabled (`ARENA_TELEMETRY=otlp`) to define 
the telemtry endpoint.

Our implementation uses OpenTelemetry (OTEL) and its Protocol (OTLP) for encoding and transport.

Default: "http://localhost:4317"
"""

OTEL_LOG_LEVEL = "OTEL_LOG_LEVEL"
"""
.. envvar:: OTEL_LOG_LEVEL

The :envvar:`OTEL_LOG_LEVEL` environment variable sets the log level used by the logger 
implementation (ArenaTelemetry) using OpenTelemetry (OTEL). 
Default: "info". 
"""

PROGRAM_STATS_UPDATE_INTERVAL_MS = "PROGRAM_STATS_UPDATE_INTERVAL_MS"
"""
.. envvar:: PROGRAM_STATS_UPDATE_INTERVAL_MS

The :envvar:`PROGRAM_STATS_UPDATE_INTERVAL_MS` environment variable defines how often program
stats are published

Default: 5000. 
"""

""" Environment variables default values """
ENV_DEFAULTS = {
  ENABLE_INTERPRETER:   'false',
  OTLP_ENDPOINT:        'http://localhost:4317',
  OTEL_LOG_LEVEL:       'info',
  PROGRAM_STATS_UPDATE_INTERVAL_MS: 5000
}

def _get_env(env_var, dft_val=None):
  """ Get value of env variable with default defined by ENV_DEFAULTS; Returns dft_val if not defined in ENV_DEFAULTS """
  return os.environ.get(env_var, ENV_DEFAULTS.get(env_var, dft_val))

def _get_arena_env():
  """Get all variables defined in this module; skip credentials, private data and imports"""
  env = {}
  skip = ( ARENA_PASSWORD, ARENA_USERNAME, 'os', 'sys' )
  for key in [ v for v in dir(sys.modules[__name__]) if not v.startswith('_') and v not in skip]:
    env[key] = os.environ.get(key)
  return env



from .attribute import Attribute

class SpeParticles(Attribute):
    """
    SpeParticles attribute class to manage its properties in the ARENA: GPU based particle systems in A-Frame.  More properties at (https://github.com/harlyq/aframe-spe-particles-component) A-Frame SPE Particles component.
    Usage: spe_particles=SpeParticles(...)
    
    :param dict acceleration: For sphere and disc distributions, only the x axis is used. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param str accelerationDistribution: Distribution of particle acceleration, for disc and sphere, only the x component will be used. if set to NONE use the 'distribution' attribute for accelerationDistribution. Allows [none, box, sphere, disc] Defaults to 'none' (optional)
    :param dict accelerationSpread: Spread of the particle's acceleration. for sphere and disc distributions, only the x axis is used. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param float activeMultiplier: Multiply the rate of particles emission, if larger than 1 then the particles will be emitted in bursts. note, very large numbers will emit all particles at once. Defaults to '1' (optional)
    :param bool affectedByFog: If true, the particles are affected by THREE js fog. Defaults to 'True' (optional)
    :param float alphaTest: Alpha values below the alphaTest threshold are considered invisible. Defaults to '0' (optional)
    :param list[float] angle: 2D rotation of the particle over the particle's lifetime, max 4 elements. Defaults to '[0]' (optional)
    :param list[float] angleSpread: Spread in angle over the particle's lifetime, max 4 elements. Defaults to '[0]' (optional)
    :param str blending: Blending mode, when drawing particles. Allows [no, normal, additive, subtractive, multiply, custom] Defaults to 'normal' (optional)
    :param list[str] color: Array of colors over the particle's lifetime, max 4 elements. Defaults to '['#fff']' (optional)
    :param list[dict] colorSpread: Spread to apply to colors, spread an array of vec3 (r g b) with 0 for no spread. note the spread will be re-applied through-out the lifetime of the particle. (optional)
    :param bool depthTest: If true, don't render a particle's pixels if something is closer in the depth buffer. Defaults to 'True' (optional)
    :param bool depthWrite: If true, particles write their depth into the depth buffer. this should be false if we use transparent particles. (optional)
    :param str direction: Make the emitter operate forward or backward in time. Allows [forward, backward] Defaults to 'forward' (optional)
    :param str distribution: Distribution for particle positions, velocities and acceleration. will be overridden by specific '...Distribution' attributes. Allows [box, sphere, disc] Defaults to 'box' (optional)
    :param float drag: Apply resistance to moving the particle, 0 is no resistance, 1 is full resistance, particle will stop moving at half of it's maxAge. Defaults to '0' (optional)
    :param float dragSpread: Spread to apply to the drag attribute. Defaults to '0' (optional)
    :param float duration: Duration of the emitter (seconds), if less than 0 then continuously emit particles. Defaults to '-1' (optional)
    :param float emitterScale: Global scaling factor for all particles from the emitter. Defaults to '100' (optional)
    :param bool enableInEditor: Enable the emitter while the editor is active (i.e. while scene is paused). (optional)
    :param bool enabled: Enable/disable the emitter. Defaults to 'True' (optional)
    :param bool frustumCulled: Enable/disable frustum culling. (optional)
    :param bool hasPerspective: If true, particles will be larger the closer they are to the camera. Defaults to 'True' (optional)
    :param float maxAge: Maximum age of a particle before reusing. Defaults to '1' (optional)
    :param float maxAgeSpread: Variance for the 'maxAge' attribute. Defaults to '0' (optional)
    :param list[float] opacity: Opacity over the particle's lifetime, max 4 elements. Defaults to '[1]' (optional)
    :param list[float] opacitySpread: Spread in opacity over the particle's lifetime, max 4 elements. Defaults to '[0]' (optional)
    :param int particleCount: Maximum number of particles for this emitter. Defaults to '100' (optional)
    :param str positionDistribution: Distribution of particle positions, disc and sphere will use the radius attributes. For box particles emit at 0,0,0, for sphere they emit on the surface of the sphere and for disc on the edge of a 2D disc on the XY plane. Allows [none, box, sphere, disc] Defaults to 'none' (optional)
    :param dict positionOffset: Fixed offset to the apply to the emitter relative to its parent entity. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param dict positionSpread: Particles are positioned within +- of these local bounds. for sphere and disc distributions only the x axis is used. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param float radius: Radius of the disc or sphere emitter (ignored for box). note radius of 0 will prevent velocity and acceleration if they use a sphere or disc distribution. Defaults to '1' (optional)
    :param dict radiusScale: Scales the emitter for sphere and disc shapes to form oblongs and ellipses. Defaults to '{'x': 1, 'y': 1, 'z': 1}' (optional)
    :param bool randomizeAcceleration: If true, re-randomize acceleration when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeAngle: If true, re-randomize angle when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeColor: If true, re-randomize colour when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeDrag: If true, re-randomize drag when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeOpacity: If true, re-randomize opacity when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizePosition: If true, re-randomize position when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeRotation: If true, re-randomize rotation when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeSize: If true, re-randomize size when re-spawning a particle, can incur a performance hit. (optional)
    :param bool randomizeVelocity: If true, re-randomize velocity when re-spawning a particle, can incur a performance hit. (optional)
    :param str relative: World relative particles move relative to the world, while local particles move relative to the emitter (i.e. if the emitter moves, all particles move with it). Allows [local, world] Defaults to 'local' (optional)
    :param float rotation: Rotation in degrees. Defaults to '0' (optional)
    :param dict rotationAxis: Local axis when using rotation. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param dict rotationAxisSpread: Variance in the axis of rotation. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param float rotationSpread: Rotation variance in degrees. Defaults to '0' (optional)
    :param bool rotationStatic: If true, the particles are fixed at their initial rotation value. if false, the particle will rotate from 0 to the 'rotation' value. (optional)
    :param list[float] size: Array of sizes over the particle's lifetime, max 4 elements. Defaults to '[1]' (optional)
    :param list[float] sizeSpread: Spread in size over the particle's lifetime, max 4 elements. Defaults to '[0]' (optional)
    :param str texture: Texture to be used for each particle, may be a spritesheet.  Examples: [blob.png, fog.png, square.png, explosion_sheet.png, fireworks_sheet.png], like path 'static/images/textures/blob.png'. (optional)
    :param int textureFrameCount: Number of frames in the spritesheet, negative numbers default to textureFrames.x * textureFrames.y. Defaults to '-1' (optional)
    :param int textureFrameLoop: Number of times the spritesheet should be looped over the lifetime of a particle. Defaults to '1' (optional)
    :param dict textureFrames: X and Y frames for a spritesheet. each particle will transition through every frame of the spritesheet over its lifetime (see textureFramesLoop). Defaults to '{'x': 1, 'y': 1}' (optional)
    :param bool useTransparency: Should the particles be rendered with transparency? Defaults to 'True' (optional)
    :param dict velocity: For sphere and disc distributions, only the x axis is used. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param str velocityDistribution: Distribution of particle velocities, for disc and sphere, only the x component will be used. if set to NONE use the 'distribution' attribute for velocityDistribution. Allows [none, box, sphere, disc] Defaults to 'none' (optional)
    :param dict velocitySpread: Variance for the velocity. Defaults to '{'x': 0, 'y': 0, 'z': 0}' (optional)
    :param float wiggle: Extra distance the particle moves over its lifetime. Defaults to '0' (optional)
    :param float wiggleSpread: +- spread for the wiggle attribute. Defaults to '0' (optional)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Particles(SpeParticles):
    """
    Alternate name for SpeParticles.
    Usage: spe_particles=Particles(...)
    """
